@@ -4,26 +4,37 @@ import json
 import os.path
 from typing import Any
 from typing import Dict
-from typing import FrozenSet
 from typing import List
 from typing import Match
 from typing import NamedTuple
 from typing import Optional
+from typing import Sequence
 from typing import Tuple
+from typing import TypeVar
+
+from identify.identify import tags_from_filename
 
 from babi._types import Protocol
 from babi.fdict import FDict
 from babi.reg import _Reg
 from babi.reg import _RegSet
 from babi.reg import ERR_REG
+from babi.reg import expand_escaped
 from babi.reg import make_reg
 from babi.reg import make_regset
 
+T = TypeVar('T')
 Scope = Tuple[str, ...]
 Regions = Tuple['Region', ...]
 Captures = Tuple[Tuple[int, '_Rule'], ...]
 
 
+def uniquely_constructed(t: T) -> T:
+    """avoid tuple.__hash__ for "singleton" constructed objects"""
+    t.__hash__ = object.__hash__  # type: ignore
+    return t
+
+
 def _split_name(s: Optional[str]) -> Tuple[str, ...]:
     if s is None:
         return ()
@@ -59,6 +70,7 @@ class _Rule(Protocol):
     def patterns(self) -> 'Tuple[_Rule, ...]': ...
 
 
+@uniquely_constructed
 class Rule(NamedTuple):
     name: Tuple[str, ...]
     match: Optional[str]
@@ -114,6 +126,10 @@ class Rule(NamedTuple):
         else:
             while_captures = ()
 
+        # some grammars (at least xml) have begin rules with no end
+        if begin is not None and end is None and while_ is None:
+            end = '$impossible^'
+
         # Using the captures key for a begin/end/while rule is short-hand for
         # giving both beginCaptures and endCaptures with same values
         if begin and end and captures:
@@ -146,24 +162,15 @@ class Rule(NamedTuple):
         )
 
 
+@uniquely_constructed
 class Grammar(NamedTuple):
     scope_name: str
-    first_line_match: Optional[_Reg]
-    file_types: FrozenSet[str]
     patterns: Tuple[_Rule, ...]
     repository: FDict[str, _Rule]
 
     @classmethod
     def from_data(cls, data: Dict[str, Any]) -> 'Grammar':
         scope_name = data['scopeName']
-        if 'firstLineMatch' in data:
-            first_line_match: Optional[_Reg] = make_reg(data['firstLineMatch'])
-        else:
-            first_line_match = None
-        if 'fileTypes' in data:
-            file_types = frozenset(data['fileTypes'])
-        else:
-            file_types = frozenset()
         patterns = tuple(Rule.from_dct(dct) for dct in data['patterns'])
         if 'repository' in data:
             repository = FDict({
@@ -173,40 +180,10 @@ class Grammar(NamedTuple):
             repository = FDict({})
         return cls(
             scope_name=scope_name,
-            first_line_match=first_line_match,
-            file_types=file_types,
             patterns=patterns,
             repository=repository,
         )
 
-    @classmethod
-    def parse(cls, filename: str) -> 'Grammar':
-        with open(filename) as f:
-            return cls.from_data(json.load(f))
-
-    @classmethod
-    def blank(cls) -> 'Grammar':
-        return cls(
-            scope_name='source.unknown',
-            first_line_match=None,
-            file_types=frozenset(),
-            patterns=(),
-            repository=FDict({}),
-        )
-
-    def matches_file(self, filename: str, first_line: str) -> bool:
-        _, ext = os.path.splitext(filename)
-        if ext.lstrip('.') in self.file_types:
-            return True
-        elif self.first_line_match is not None:
-            return bool(
-                self.first_line_match.match(
-                    first_line, 0, first_line=True, boundary=True,
-                ),
-            )
-        else:
-            return False
-
 
 class Region(NamedTuple):
     start: int
@@ -369,6 +346,7 @@ def _do_regset(
     return state, match.end(), boundary, tuple(ret)
 
 
+@uniquely_constructed
 class PatternRule(NamedTuple):
     name: Tuple[str, ...]
     regset: _RegSet
@@ -395,6 +373,7 @@ class PatternRule(NamedTuple):
         return _do_regset(idx, match, self, compiler, state, pos)
 
 
+@uniquely_constructed
 class MatchRule(NamedTuple):
     name: Tuple[str, ...]
     captures: Captures
@@ -420,6 +399,7 @@ class MatchRule(NamedTuple):
         raise AssertionError(f'unreachable {self}')
 
 
+@uniquely_constructed
 class EndRule(NamedTuple):
     name: Tuple[str, ...]
     content_name: Tuple[str, ...]
@@ -439,7 +419,7 @@ class EndRule(NamedTuple):
         next_scope = scope + self.content_name
 
         boundary = match.end() == len(match.string)
-        reg = make_reg(match.expand(self.end))
+        reg = make_reg(expand_escaped(match, self.end))
         state = state.push(Entry(next_scope, self, reg, boundary))
         regions = _captures(compiler, scope, match, self.begin_captures)
         return state, True, regions
@@ -480,6 +460,7 @@ class EndRule(NamedTuple):
                 return _do_regset(idx, match, self, compiler, state, pos)
 
 
+@uniquely_constructed
 class WhileRule(NamedTuple):
     name: Tuple[str, ...]
     content_name: Tuple[str, ...]
@@ -499,7 +480,7 @@ class WhileRule(NamedTuple):
         next_scope = scope + self.content_name
 
         boundary = match.end() == len(match.string)
-        reg = make_reg(match.expand(self.while_))
+        reg = make_reg(expand_escaped(match, self.while_))
         state = state.push_while(self, Entry(next_scope, self, reg, boundary))
         regions = _captures(compiler, scope, match, self.begin_captures)
         return state, True, regions
@@ -534,7 +515,7 @@ class WhileRule(NamedTuple):
 
 
 class Compiler:
-    def __init__(self, grammar: Grammar, grammars: Dict[str, Grammar]) -> None:
+    def __init__(self, grammar: Grammar, grammars: 'Grammars') -> None:
         self._root_scope = grammar.scope_name
         self._grammars = grammars
         self._rule_to_grammar: Dict[_Rule, Grammar] = {}
@@ -555,14 +536,17 @@ class Compiler:
         if s == '$self':
             return self._patterns(grammar, grammar.patterns)
         elif s == '$base':
-            return self._include(self._grammars[self._root_scope], '$self')
+            grammar = self._grammars.grammar_for_scope(self._root_scope)
+            return self._include(grammar, '$self')
         elif s.startswith('#'):
             return self._patterns(grammar, (grammar.repository[s[1:]],))
         elif '#' not in s:
-            return self._include(self._grammars[s], '$self')
+            grammar = self._grammars.grammar_for_scope(s)
+            return self._include(grammar, '$self')
         else:
             scope, _, s = s.partition('#')
-            return self._include(self._grammars[scope], f'#{s}')
+            grammar = self._grammars.grammar_for_scope(scope)
+            return self._include(grammar, f'#{s}')
 
     @functools.lru_cache(maxsize=None)
     def _patterns(
@@ -643,49 +627,64 @@ class Compiler:
 
 
 class Grammars:
-    def __init__(self, grammars: List[Grammar]) -> None:
-        self.grammars = {grammar.scope_name: grammar for grammar in grammars}
-        self._compilers: Dict[Grammar, Compiler] = {}
+    def __init__(self, grammars: Sequence[Dict[str, Any]]) -> None:
+        self._raw = {grammar['scopeName']: grammar for grammar in grammars}
+        self._find_scope = [
+            (
+                frozenset(grammar.get('fileTypes', ())),
+                make_reg(grammar.get('firstLineMatch', '$impossible^')),
+                grammar['scopeName'],
+            )
+            for grammar in grammars
+        ]
+        self._parsed: Dict[str, Grammar] = {}
+        self._compilers: Dict[str, Compiler] = {}
 
     @classmethod
     def from_syntax_dir(cls, syntax_dir: str) -> 'Grammars':
-        grammars = [Grammar.blank()]
+        grammars = [{'scopeName': 'source.unknown', 'patterns': []}]
         if os.path.exists(syntax_dir):
-            grammars.extend(
-                Grammar.parse(os.path.join(syntax_dir, filename))
-                for filename in os.listdir(syntax_dir)
-            )
+            for filename in os.listdir(syntax_dir):
+                with open(os.path.join(syntax_dir, filename)) as f:
+                    grammars.append(json.load(f))
         return cls(grammars)
 
-    def _compiler_for_grammar(self, grammar: Grammar) -> Compiler:
+    def grammar_for_scope(self, scope: str) -> Grammar:
         with contextlib.suppress(KeyError):
-            return self._compilers[grammar]
+            return self._parsed[scope]
 
-        ret = self._compilers[grammar] = Compiler(grammar, self.grammars)
+        ret = self._parsed[scope] = Grammar.from_data(self._raw[scope])
         return ret
 
     def compiler_for_scope(self, scope: str) -> Compiler:
-        return self._compiler_for_grammar(self.grammars[scope])
+        with contextlib.suppress(KeyError):
+            return self._compilers[scope]
+
+        grammar = self.grammar_for_scope(scope)
+        ret = self._compilers[scope] = Compiler(grammar, self)
+        return ret
 
     def blank_compiler(self) -> Compiler:
         return self.compiler_for_scope('source.unknown')
 
-    def compiler_for_file(self, filename: str) -> Compiler:
-        if os.path.exists(filename):
-            with open(filename) as f:
-                first_line = next(f, '')
+    def compiler_for_file(self, filename: str, first_line: str) -> Compiler:
+        for tag in tags_from_filename(filename) - {'text'}:
+            with contextlib.suppress(KeyError):
+                return self.compiler_for_scope(f'source.{tag}')
+
+        _, _, ext = os.path.basename(filename).rpartition('.')
+        for extensions, first_line_match, scope_name in self._find_scope:
+            if (
+                    ext in extensions or
+                    first_line_match.match(
+                        first_line, 0, first_line=True, boundary=True,
+                    )
+            ):
+                return self.compiler_for_scope(scope_name)
         else:
-            first_line = ''
-        for grammar in self.grammars.values():
-            if grammar.matches_file(filename, first_line):
-                break
-        else:
-            grammar = self.grammars['source.unknown']
-
-        return self._compiler_for_grammar(grammar)
+            return self.compiler_for_scope('source.unknown')
 
 
-@functools.lru_cache(maxsize=None)
 def highlight_line(
         compiler: 'Compiler',
         state: State,
