@@ -6,7 +6,6 @@ import os
 import re
 import signal
 import sys
-from typing import Callable
 from typing import Generator
 from typing import List
 from typing import NamedTuple
@@ -21,15 +20,19 @@ from babi.file import File
 from babi.file import get_lines
 from babi.history import History
 from babi.hl.syntax import Syntax
-from babi.hl.trailing_whitespace import TrailingWhitespace
 from babi.margin import Margin
 from babi.perf import Perf
 from babi.prompt import Prompt
 from babi.prompt import PromptResult
 from babi.status import Status
 
-VERSION_STR = 'babi v0'
-EditResult = enum.Enum('EditResult', 'EXIT NEXT PREV')
+if sys.version_info >= (3, 8):  # pragma: no cover (py38+)
+    import importlib.metadata as importlib_metadata
+else:  # pragma: no cover (<py38)
+    import importlib_metadata
+
+VERSION_STR = f'babi v{importlib_metadata.version("babi")}'
+EditResult = enum.Enum('EditResult', 'EXIT NEXT PREV OPEN')
 
 # TODO: find a place to populate these, surely there's a database somewhere
 SEQUENCE_KEYNAME = {
@@ -73,12 +76,12 @@ class Screen:
             perf: Perf,
     ) -> None:
         self.stdscr = stdscr
-        color_manager = ColorManager.make()
-        hl_factories = (
-            Syntax.from_screen(stdscr, color_manager),
-            TrailingWhitespace(color_manager),
-        )
-        self.files = [File(f, hl_factories) for f in filenames]
+        self.color_manager = ColorManager.make()
+        self.hl_factories = (Syntax.from_screen(stdscr, self.color_manager),)
+        self.files = [
+            File(filename, self.color_manager, self.hl_factories)
+            for filename in filenames
+        ]
         self.i = 0
         self.history = History()
         self.perf = perf
@@ -86,7 +89,6 @@ class Screen:
         self.margin = Margin.from_current_screen()
         self.cut_buffer: Tuple[str, ...] = ()
         self.cut_selection = False
-        self._resize_cb: Optional[Callable[[], None]] = None
         self._buffered_input: Union[int, str, None] = None
 
     @property
@@ -220,30 +222,46 @@ class Screen:
         self.file.draw(self.stdscr, self.margin)
         self.status.draw(self.stdscr, self.margin)
 
-    @contextlib.contextmanager
-    def resize_cb(self, f: Callable[[], None]) -> Generator[None, None, None]:
-        assert self._resize_cb is None, self._resize_cb
-        self._resize_cb = f
-        try:
-            yield
-        finally:
-            self._resize_cb = None
-
     def resize(self) -> None:
         curses.update_lines_cols()
         self.margin = Margin.from_current_screen()
         self.file.scroll_screen_if_needed(self.margin)
         self.draw()
-        if self._resize_cb is not None:
-            self._resize_cb()
 
-    def quick_prompt(self, prompt: str, opts: str) -> Union[str, PromptResult]:
+    def quick_prompt(
+            self,
+            prompt: str,
+            opt_strs: Tuple[str, ...],
+    ) -> Union[str, PromptResult]:
+        opts = [opt[0] for opt in opt_strs]
         while True:
-            s = prompt.ljust(curses.COLS)
-            if len(s) > curses.COLS:
-                s = f'{s[:curses.COLS - 1]}…'
-            self.stdscr.insstr(curses.LINES - 1, 0, s, curses.A_REVERSE)
-            x = min(curses.COLS - 1, len(prompt) + 1)
+            x = 0
+
+            def _write(s: str, *, attr: int = curses.A_REVERSE) -> None:
+                nonlocal x
+
+                if x >= curses.COLS:
+                    return
+                self.stdscr.insstr(curses.LINES - 1, x, s, attr)
+                x += len(s)
+
+            _write(prompt)
+            _write(' [')
+            for i, opt_str in enumerate(opt_strs):
+                _write(opt_str[0], attr=curses.A_REVERSE | curses.A_BOLD)
+                _write(opt_str[1:])
+                if i != len(opt_strs) - 1:
+                    _write(', ')
+            _write(']?')
+
+            if x < curses.COLS - 1:
+                s = ' ' * (curses.COLS - x)
+                self.stdscr.insstr(curses.LINES - 1, x, s, curses.A_REVERSE)
+                x += 1
+            else:
+                x = curses.COLS - 1
+                self.stdscr.insstr(curses.LINES - 1, x, '…', curses.A_REVERSE)
+
             self.stdscr.move(curses.LINES - 1, x)
 
             key = self.get_char()
@@ -306,7 +324,7 @@ class Screen:
         self.status.update(f'{line}, {col} (of {line_count} {lines_word})')
 
     def cut(self) -> None:
-        if self.file.select_start:
+        if self.file.selection.start:
             self.cut_buffer = self.file.cut_selection(self.margin)
             self.cut_selection = True
         else:
@@ -342,6 +360,7 @@ class Screen:
             to_stack.append(action.apply(self.file))
             self.file.scroll_screen_if_needed(self.margin)
             self.status.update(f'{op}: {action.name}')
+            self.file.selection.clear()
 
     def undo(self) -> None:
         self._undo_redo('undo', self.file.undo_stack, self.file.redo_stack)
@@ -373,7 +392,7 @@ class Screen:
             self.save()
             return EditResult.EXIT
         elif response == ':sort':
-            if self.file.select_start:
+            if self.file.selection.start:
                 self.file.sort_selection(self.margin)
             else:
                 self.file.sort(self.margin)
@@ -436,10 +455,19 @@ class Screen:
             self.file.filename = response
             return self.save()
 
+    def open_file(self) -> Optional[EditResult]:
+        response = self.prompt('enter filename', history='open')
+        if response is not PromptResult.CANCELLED:
+            opened = File(response, self.color_manager, self.hl_factories)
+            self.files.append(opened)
+            return EditResult.OPEN
+        else:
+            return None
+
     def quit_save_modified(self) -> Optional[EditResult]:
         if self.file.modified:
             response = self.quick_prompt(
-                'file is modified - save [y(es), n(o)]?', 'yn',
+                'file is modified - save', ('yes', 'no'),
             )
             if response == 'y':
                 if self.save_filename() is not PromptResult.CANCELLED:
@@ -473,6 +501,7 @@ class Screen:
         b'^S': save,
         b'^O': save_filename,
         b'^X': quit_save_modified,
+        b'^P': open_file,
         b'kLFT3': lambda screen: EditResult.PREV,
         b'kRIT3': lambda screen: EditResult.NEXT,
         b'^Z': background,
